# Generated by Django 5.2 on 2026-03-11 00:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentikate", "0004_device_client_device"),
    ]

    operations = [
        migrations.AlterField(
            model_name="client",
            name="client_id",
            field=models.CharField(max_length=2000),
        ),
    ]
